%%%-------------------------------------------------------------------
%%% File    : berkeley_db_replication_group_gen_event_handler.erl
%%% Author  : Chris Newcombe
%%% Description : 
%%%
%%%-------------------------------------------------------------------
-module(berkeley_db_replication_group_gen_event_handler).

-export([init/1, handle_event/2, terminate/2]).

%%
%% This handles events from ALL port_coordinator replication_gen_event instances
%% in the replication group.
%%
%% We just forward all events (appropriately tagged) as messates to the 
%% replication_group instance that installed the handler.
%%

init({forward_events_as_messages_to, _, _} = State) ->
    io:format("~w:successfully subscribed (State=~w)~n", [?MODULE, State]),
    {ok, State}.


handle_event(Event, {forward_events_as_messages_to, MsgTag, ForwardToPid} = State) ->
    io:format("~w:handle_event received (Event=~w, State=~w)~n", [?MODULE, Event, State]),
    ForwardToPid ! {MsgTag, Event},
    {ok, State}.


%% If the event handler is deleted due to a call to gen_event:delete_handler, gen_event:swap_handler/3 or gen_event:swap_sup_handler/3, 
%% then Arg is the Args argument of that function call. 
%%
%% Arg={stop,Reason} if the event handler has a supervised connection to a process which has terminated with reason Reason. 
%%
%% Arg=stop if the event handler is deleted because the event manager is terminating. 
%%
%% Arg=remove_handler if the event handler is deleted because another callback function has returned remove_handler or {remove_handler,Reply}. 
%%
%% Arg={error,Term} if the event handler is deleted because a callback function returned an unexpected value Term, or Arg={error,{'EXIT',Reason}} if a callback function failed. 
terminate(Args, State) ->
    io:format("~w:terminate received (Args=~w, State=~w)~n", [?MODULE, Args, State]),
    ok.
