%%%-------------------------------------------------------------------------
%%% File    : ethbridge.erl
%%% Purpose : Top-level Ethernet bridge application module.
%%% Created : 01 Aug 2002 by Scott Lystig Fritchie <slfritchie@snookles.com>
%%%-------------------------------------------------------------------------

-module(ethbridge).
-compile([verbose, report_errors, report_warnings, trace, debug_info]).

-behaviour(application).

%% application callbacks
-export([start/0, start/2, stop/1]).

%%%----------------------------------------------------------------------
%%% Callback functions from application
%%%----------------------------------------------------------------------

%%%----------------------------------------------------------------------
%%% This isn't a full-blowing application.  If this were a full
%%% application, more stuff would be here.  Besides, I don't know
%%% if it's possible to have multiple applications running
%%% simultaneously but with different names ... which is what I'm
%%% interested in.
%%%----------------------------------------------------------------------

%%----------------------------------------------------------------------
%% Func: start/2
%% Returns: {ok, Pid}        |
%%          {ok, Pid, State} |
%%          {error, Reason}   
%%----------------------------------------------------------------------
start() ->
    start(xxxwhocares, [{bridge_name, "foofoo"}, {foo, foo}, {bar, bar},
		        {interfaces_from, env}]).
start(Type, StartArgs) ->
    io:format("XXX ~w:start Type = ~w, StartArgs = ~w\n", [?MODULE, Type, StartArgs]),
    %%
    %% Application environment values: fetch via application:get_all_env et al.
    %% Sources:
    %%    1. 'env' portion of ethbridge.app file
    %%    2. from a .config file via "-config filename" on command line
    %%    3. from "-ethbridge keyX valX keyY valY ..."
    %%
    io:format("XXX ~w:start get_all_env() = ~w\n", [?MODULE,
						    application:get_all_env(ethbridge)]),
    case lists:keysearch(bridge_name, 1, StartArgs) of
	{value, {bridge_name, BName}} ->
	    CName = list_to_atom(BName ++ "_center"),
	    ISName = list_to_atom(BName ++ "_iface_sup"),
	    ethbridge_sup:start_link(StartArgs ++ [{center_name, CName},
						   {iface_sup_name, ISName}]);
	_    -> {error, "Must include {bridge_name, String} in options"}
    end.
    

%%----------------------------------------------------------------------
%% Func: stop/1
%% Returns: any 
%%----------------------------------------------------------------------
stop(State) ->
    ok.

%%%----------------------------------------------------------------------
%%% Internal functions
%%%----------------------------------------------------------------------
