%%%----------------------------------------------------------------------
%%% File    : ethbridge_sup.erl
%%% Purpose : Top-level supervisor for Ethernet bridge.
%%% Created : 01 Aug 2002 by Scott Lystig Fritchie <slfritchie@snookles.com>
%%%----------------------------------------------------------------------

-module(ethbridge_sup).
-compile([verbose, report_errors, report_warnings, trace, debug_info]).

-behaviour(supervisor).

%% External exports
-export([start_link/0, start_link/1]).

%% supervisor callbacks
-export([init/1]).

%%%----------------------------------------------------------------------
%%% API
%%%----------------------------------------------------------------------
start_link() ->
    io:format("XXX ~w:start_link()\n", [?MODULE]),
    {error, "Must include StartArgs list"}.
start_link(StartArgs) ->
    io:format("XXX ~w:start_link ArgList = ~w\n", [?MODULE, StartArgs]),
    case lists:keysearch(bridge_name, 1, StartArgs) of
	{value, {bridge_name, BName}} ->
	    BSName = list_to_atom(BName ++ "_sup"),
	    supervisor:start_link({local, BSName}, ethbridge_sup, StartArgs);
	_ ->
	    {error, {bridge_name, not_found}}
    end.

%%%----------------------------------------------------------------------
%%% Callback functions from supervisor
%%%----------------------------------------------------------------------

%%----------------------------------------------------------------------
%% Func: init/1
%% Returns: {ok,  {SupFlags,  [ChildSpec]}} |
%%          ignore                          |
%%          {error, Reason}   
%%
%% The cd_sup supervisor will use a one_for_all strategy: if a child dies,
%% all remaining children are killed before restarting the children.
%%----------------------------------------------------------------------

init(ArgList) ->
    io:format("XXX ~w:init ArgList = ~w\n", [?MODULE, ArgList]),
    %% Child_spec = {IdName, {M, F, A},
    %%               Restart, Shutdown_time, Type, Modules_used}
    Center = {center, {ethbr_center, start_link, [ArgList]},
		     permanent, 2000, worker, [ethbr_center]},
    IFaceSupervisor =  {ethbr_iface_sup,
			{ethbr_iface_sup, start_link, [ArgList]},
			permanent, 5000, supervisor, [ ethbr_iface_sup ]},
    {ok, {{rest_for_one, 2, 3600}, [
				    Center,
				    IFaceSupervisor
				   ]}}.

%%%----------------------------------------------------------------------
%%% Internal functions
%%%----------------------------------------------------------------------

%% None.
