%%%----------------------------------------------------------------------
%%% File    : libnet_drv.erl
%%% Summary : Erlang driver for the libnet
%%%
%%%
%%% NOTICE: This file was generated by the tools of the Erlang Driver
%%%         toolkit.  Do not edit this file by hand unless you know
%%%         what you're doing!
%%%
%%% Copyright (c) 2002, Scott Lystig Fritchie.  All rights reserved.
%%% See the file "LICENSE" at the top of the source distribution for
%%% full license terms.
%%%
%%%----------------------------------------------------------------------

-module(libnet_drv).
-include("libnet_drv.hrl").

%% Xref with erl_driver_tk.h's PIPE_DRIVER_TERM_* values
-define(T_NIL, 0).
-define(T_ATOM, 1).
-define(T_PORT, 2).
-define(T_INT, 3).
-define(T_TUPLE, 4).
-define(T_BINARY, 5).
-define(T_STRING, 6).
-define(T_LIST, 7).

%% External exports
-export([start/0, start_pipe/0]).
-export([shutdown/1]).
-export([debug/2]).
-export([
         null/1, 
         libnet_open_raw_sock/2, 
         libnet_close_raw_sock/2, 
         my_libnet_select_device_any/1, 
         libnet_select_device/2, 
         libnet_open_link_interface/2, 
         libnet_close_link_interface/2, 
         libnet_write_ip/3, 
         libnet_write_link_layer/4, 
         libnet_do_checksum/4, 
         libnet_ip_check/2, 
         test_send_ping/4, 
         test_make_icmp_echo_no_checksum/3
        ]).

start() ->
    {ok, Path} = load_path(?DRV_NAME ++ ".so"),
    erl_ddll:start(),
    ok = erl_ddll:load_driver(Path, ?DRV_NAME),
    case open_port({spawn, ?DRV_NAME}, []) of
        P when port(P) ->
            {ok, P};
        Err ->
            Err
    end.

start_pipe() ->
    {ok, PipeMain} = load_path("pipe-main"),
    {ok, ShLib} = load_path("./libnet_drv.so"),
    Cmd = PipeMain ++ "/pipe-main " ++ ShLib ++ "/libnet_drv.so",
    case open_port({spawn, Cmd}, [exit_status, binary, use_stdio, {packet, 4}]) of
        P when port(P) ->
            {ok, P};
        Err ->
            Err
    end.

shutdown(Port) when port(Port) ->
    catch erlang:port_close(Port),
    %% I was under the impression you'd always get a message sent to
    %% you in this case, so this receive is to keep your mailbox from
    %% getting cluttered.  Hrm, well, sometimes the message does
    %% not arrive at all!
    receive
        {'EXIT', Port, normal} -> {ok, normal};
        {'EXIT', Port, Err}    -> {error, Err}
    after 0                    -> {ok, normall} % XXX is 0 too small?
        
    end.

debug(Port, Flags) when port(Port), integer(Flags) ->
    case catch erlang:port_command(Port, <<?_DEBUG, Flags:32>>) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX too drastic?
    end.

null(Port
        ) when port(Port) -> % TODO: Add additional constraints here
    IOList_____ = <<?_NULL>>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

libnet_open_raw_sock(Port,
     Protocol
        ) when port(Port) -> % TODO: Add additional constraints here
    IOList_____ = <<?_LIBNET_OPEN_RAW_SOCK,
          Protocol:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

libnet_close_raw_sock(Port,
     Fd
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_raw_sock, FdIndex} = Fd,
    IOList_____ = <<?_LIBNET_CLOSE_RAW_SOCK,
            FdIndex:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

my_libnet_select_device_any(Port
        ) when port(Port) -> % TODO: Add additional constraints here
    IOList_____ = <<?_MY_LIBNET_SELECT_DEVICE_ANY>>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

libnet_select_device(Port,
     Device
        ) when port(Port) -> % TODO: Add additional constraints here
    {DeviceBinOrList, DeviceLen} = serialize_contiguously(Device, 1),
    IOList_____ = [ <<?_LIBNET_SELECT_DEVICE,
            DeviceLen:32/integer>>,             % I/O list length
          DeviceBinOrList,
          <<
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

libnet_open_link_interface(Port,
     Device
        ) when port(Port) -> % TODO: Add additional constraints here
    {DeviceBinOrList, DeviceLen} = serialize_contiguously(Device, 1),
    IOList_____ = [ <<?_LIBNET_OPEN_LINK_INTERFACE,
            DeviceLen:32/integer>>,             % I/O list length
          DeviceBinOrList,
          <<
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

libnet_close_link_interface(Port,
     L
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_link_int, LIndex} = L,
    IOList_____ = <<?_LIBNET_CLOSE_LINK_INTERFACE,
            LIndex:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

libnet_write_ip(Port,
     Fd, 
     Buf
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_raw_sock, FdIndex} = Fd,
    {BufBinOrList, BufLen} = serialize_contiguously(Buf, 0),
    IOList_____ = [ <<?_LIBNET_WRITE_IP,
            FdIndex:32/integer, 
            BufLen:32/integer>>,                % I/O list length
          BufBinOrList,
          <<
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

libnet_write_link_layer(Port,
     L, 
     Device, 
     Buf
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_link_int, LIndex} = L,
    {DeviceBinOrList, DeviceLen} = serialize_contiguously(Device, 1),
    {BufBinOrList, BufLen} = serialize_contiguously(Buf, 0),
    IOList_____ = [ <<?_LIBNET_WRITE_LINK_LAYER,
            LIndex:32/integer, 
            DeviceLen:32/integer>>,             % I/O list length
          DeviceBinOrList,
          <<
            BufLen:32/integer>>,                % I/O list length
          BufBinOrList,
          <<
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

libnet_do_checksum(Port,
     Buf, 
     Protocol, 
     Len
        ) when port(Port) -> % TODO: Add additional constraints here
    {BufBinOrList, BufLen} = serialize_contiguously(Buf, 0),
    IOList_____ = [ <<?_LIBNET_DO_CHECKSUM,
            BufLen:32/integer>>,                % I/O list length
          BufBinOrList,
          <<
          Protocol:32/integer, 
          Len:32/integer
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

libnet_ip_check(Port,
     Buf_U_Short
        ) when port(Port) -> % TODO: Add additional constraints here
    {Buf_U_ShortBinOrList, Buf_U_ShortLen} = serialize_contiguously(Buf_U_Short, 0),
    IOList_____ = [ <<?_LIBNET_IP_CHECK,
            Buf_U_ShortLen:32/integer>>,        % I/O list length
          Buf_U_ShortBinOrList,
          <<
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

test_send_ping(Port,
     Src, 
     Dst, 
     Fd
        ) when port(Port) -> % TODO: Add additional constraints here
    {SrcBinOrList, SrcLen} = serialize_contiguously(Src, 1),
    {DstBinOrList, DstLen} = serialize_contiguously(Dst, 1),
    {valmap_raw_sock, FdIndex} = Fd,
    IOList_____ = [ <<?_TEST_SEND_PING,
            SrcLen:32/integer>>,                % I/O list length
          SrcBinOrList,
          <<
            DstLen:32/integer>>,                % I/O list length
          DstBinOrList,
          <<
            FdIndex:32/integer
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

test_make_icmp_echo_no_checksum(Port,
     Src, 
     Dst
        ) when port(Port) -> % TODO: Add additional constraints here
    {SrcBinOrList, SrcLen} = serialize_contiguously(Src, 1),
    {DstBinOrList, DstLen} = serialize_contiguously(Dst, 1),
    IOList_____ = [ <<?_TEST_MAKE_ICMP_ECHO_NO_CHECKSUM,
            SrcLen:32/integer>>,                % I/O list length
          SrcBinOrList,
          <<
            DstLen:32/integer>>,                % I/O list length
          DstBinOrList,
          <<
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

    
%%%
%%% Internal functions.
%%%

load_path(File) ->
    case lists:filter(fun(D) ->
                              case file:read_file_info(D ++ "/" ++ File) of
                                  {ok, _} -> true;
                                  _ -> false
                              end
                      end, code:get_path()) of
        [Dir|_] ->
            {ok, Dir};
        [] ->
            io:format("Error: ~s not found in code path\n", [File]),
            {error, enoent}
    end.

%%%
%%% Note that an 'xtra_return' that only returns one item in its
%%% tuple will return {Port, ok, {Thingie}}, so we'll return
%%% {ok, {Thingie}}, which is *sooooooo* maddening because I keep
%%% forgetting the extra tuple wrapper.  So, if there's only one
%%% thingie in the return tuple, strip it off: {ok, Thingie}
%%%

get_port_reply(Port) when port(Port) ->
    receive
        {Port, ok} = T -> proc_reply(T);
        {Port, ok, {M}} = T -> proc_reply(T);
        {Port, ok, M} = T -> proc_reply(T);
        {Port, error, {Reason}} = T -> proc_reply(T);
        {Port, error, Reason} = T -> proc_reply(T);
        %% Pipe driver messages
        {Port, {data, Bytes}} -> proc_reply(pipedrv_deser(Port, Bytes));
        {'EXIT', Port, Reason} -> throw({port_error, Reason});  % XXX too drastic?
        {Port, Reason} -> throw({port_error, Reason})   % XXX too drastic?
    end.

%% This function exists to provide consistency of replies 
%% given by linked-in and pipe drivers.  The "receive" statement
%% in get_port_reply/1 is specific because we want it to be
%% very selective about what it will grab out of the mailbox.
proc_reply({Port, ok}) when port(Port) ->
    ok;
proc_reply({Port, ok, {M}}) when port(Port) ->
    {ok, M};
proc_reply({Port, ok, M}) when port(Port) ->
    {ok, M};
proc_reply({Port, error, {Reason}}) when port(Port) ->
    {error, Reason};
proc_reply({Port, error, Reason}) when port(Port) ->
    {error, Reason}.


%%% io_list_len() is an extremely useful function.  BEAM has got this
%%% implemented quite efficiently in C.  It would be *fabulous* to be able
%%% to use it from Erlang via a BIF.

io_list_len(B) when binary(B) -> {B, size(B)};
io_list_len(L) when list(L) -> io_list_len(L, 0).
io_list_len([H|T], N) ->
    if
        H >= 0, H =< 255 -> io_list_len(T, N+1);
        list(H) -> io_list_len(T, io_list_len(H,N));
        binary(H) -> io_list_len(T, size(H) + N);
        true -> throw({error, partial_len, N})
    end;
io_list_len(H, N) when binary(H) -> 
    size(H) + N;
io_list_len([], N) -> 
    N.

%%% We need to make the binary thing we're passing in contiguous
%%% because the C function we're calling is expecting a single
%%% contiguous buffer.  If IOList is ["Hello, ", <<"World">>, "!"],
%%% that binary in the middle element will end up with the argument
%%% spanning three parts of an ErlIOVec.  If that happens, then we'd
%%% have to have the driver do the dirty work of putting the argument
%%% into a single contiguous buffer.
%%%
%%% Frankly, we're lazy, and this code is short and won't be much
%%% slower than doing it in C.

%%% 2nd arg: if 1, NUL-terminate the IOList

serialize_contiguously(B, 0) when binary(B) ->
    {B, size(B)};
serialize_contiguously([B], 0) when binary(B) ->
    {B, size(B)};
serialize_contiguously(IOList, 1) ->
    serialize_contiguously([IOList, 0], 0);
serialize_contiguously(IOList, 0) ->
    B = list_to_binary(IOList),
    {B, size(B)}.


%% pipedrv_deser/2 -- Deserialize the term that the pipe driver is
%% is returning to Erlang.  The pipe driver doesn't know it's a pipe
%% driver, it thinks it's a linked-in driver, so it tries to return
%% an arbitrary Erlang term to us.  The pipe-main program is sneaky:
%% it has a driver_output_term() function that serializes the term
%% that the driver built.  With the help of a list-as-stack, we
%% deserialize that term.

pipedrv_deser(Port, B) ->
    pipedrv_deser(Port, B, []).

pipedrv_deser(Port, <<>>, []) ->
    throw(icky_i_think);
pipedrv_deser(Port, <<>>, [T]) ->
    T;
pipedrv_deser(Port, <<?T_NIL:8, Rest/binary>>, Stack) ->
    pipedrv_deser(Port, Rest, [foo___foo_nil___|Stack]);
pipedrv_deser(Port, <<?T_ATOM:8, Len:8, Rest/binary>>, Stack) ->
    <<A:Len/binary, Rest2/binary>> = Rest,
    pipedrv_deser(Port, Rest2, [list_to_atom(binary_to_list(A))|Stack]);
pipedrv_deser(Port, <<?T_PORT:8, P:32/unsigned, Rest/binary>>, Stack) ->
    %% The pipe driver tried sending us a port, but it cannot know what
    %% port ID was assigned to this port, so we'll assume it is Port.
    pipedrv_deser(Port, Rest, [Port|Stack]);
pipedrv_deser(Port, <<?T_INT:8, I:32/signed, Rest/binary>>, Stack) ->
    pipedrv_deser(Port, Rest, [I|Stack]);
pipedrv_deser(Port, <<?T_TUPLE:8, N:8, Rest/binary>>, Stack) ->
    {L, NewStack} = popN(N, Stack),
    pipedrv_deser(Port, Rest, [list_to_tuple(L)|NewStack]);
pipedrv_deser(Port, <<?T_LIST:8, N:32, Rest/binary>>, Stack) ->
    {L, NewStack} = popN(N, Stack),
    pipedrv_deser(Port, Rest, [L|NewStack]);
pipedrv_deser(Port, <<?T_BINARY:8, Len:32/signed, Rest/binary>>, Stack) ->
    <<Bin:Len/binary, Rest2/binary>> = Rest,
    pipedrv_deser(Port, Rest2, [Bin|Stack]);
pipedrv_deser(Port, <<?T_STRING:8, Len:32/signed, Rest/binary>>, Stack) ->
    <<Bin:Len/binary, Rest2/binary>> = Rest,
    pipedrv_deser(Port, Rest2, [binary_to_list(Bin)|Stack]);
pipedrv_deser(Port, X, Y) ->
    throw({bah, X, Y}).

popN(N, Stack) ->
    popN(N, Stack, []).
popN(0, Stack, Acc) ->
    {Acc, Stack};
popN(N, [foo___foo_nil___|T], Acc) ->
    %% This is the nonsense we put on the stack to represent NIL.  Ignore it.
    popN(N - 1, T, Acc);
popN(N, [H|T], Acc) ->
    popN(N - 1, T, [H|Acc]).


%%%
%%% Begin code included via <custom_erl> tags
%%%


