%%%----------------------------------------------------------------------
%%% File    : libpcap_drv.erl
%%% Summary : Erlang driver for the libpcap
%%%
%%%
%%% NOTICE: This file was generated by the tools of the Erlang Driver
%%%         toolkit.  Do not edit this file by hand unless you know
%%%         what you're doing!
%%%
%%% Copyright (c) 2002, Scott Lystig Fritchie.  All rights reserved.
%%% See the file "LICENSE" at the top of the source distribution for
%%% full license terms.
%%%
%%%----------------------------------------------------------------------

-module(libpcap_drv).
-include("libpcap_drv.hrl").

%% Xref with erl_driver_tk.h's PIPE_DRIVER_TERM_* values
-define(T_NIL, 0).
-define(T_ATOM, 1).
-define(T_PORT, 2).
-define(T_INT, 3).
-define(T_TUPLE, 4).
-define(T_BINARY, 5).
-define(T_STRING, 6).
-define(T_LIST, 7).

%% External exports
-export([start/0, start_pipe/0]).
-export([shutdown/1]).
-export([debug/2]).
-export([
         null/1, 
         pcap_open_live/5, 
         pcap_open_offline/2, 
         my_set_immediate/3, 
         my_set_see_sent/3, 
         pcap_setnonblock/3, 
         pcap_getnonblock/2, 
         pcap_findalldevs/1, 
         pcap_lookupdev/1, 
         pcap_lookupnet/2, 
         my_pcap_compile/5, 
         pcap_setfilter/3, 
         my_pcap_freecode/2, 
         pcap_next/2, 
         pcap_datalink/2, 
         pcap_snapshot/2, 
         pcap_is_swapped/2, 
         pcap_major_version/2, 
         pcap_minor_version/2, 
         pcap_stats/2, 
         pcap_fileno/2, 
         pcap_geterr/2, 
         pcap_close/2
        ]).

start() ->
    {ok, Path} = load_path(?DRV_NAME ++ ".so"),
    erl_ddll:start(),
    ok = erl_ddll:load_driver(Path, ?DRV_NAME),
    case open_port({spawn, ?DRV_NAME}, []) of
        P when port(P) ->
            {ok, P};
        Err ->
            Err
    end.

start_pipe() ->
    {ok, PipeMain} = load_path("pipe-main"),
    {ok, ShLib} = load_path("./libpcap_drv.so"),
    Cmd = PipeMain ++ "/pipe-main " ++ ShLib ++ "/libpcap_drv.so",
    case open_port({spawn, Cmd}, [exit_status, binary, use_stdio, {packet, 4}]) of
        P when port(P) ->
            {ok, P};
        Err ->
            Err
    end.

shutdown(Port) when port(Port) ->
    catch erlang:port_close(Port),
    %% I was under the impression you'd always get a message sent to
    %% you in this case, so this receive is to keep your mailbox from
    %% getting cluttered.  Hrm, well, sometimes the message does
    %% not arrive at all!
    receive
        {'EXIT', Port, normal} -> {ok, normal};
        {'EXIT', Port, Err}    -> {error, Err}
    after 0                    -> {ok, normall} % XXX is 0 too small?
        
    end.

debug(Port, Flags) when port(Port), integer(Flags) ->
    case catch erlang:port_command(Port, <<?_DEBUG, Flags:32>>) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX too drastic?
    end.

null(Port
        ) when port(Port) -> % TODO: Add additional constraints here
    IOList_____ = <<?_NULL>>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

pcap_open_live(Port,
     Device, 
     Snaplen, 
     Promisc, 
     To_Ms
        ) when port(Port) -> % TODO: Add additional constraints here
    {DeviceBinOrList, DeviceLen} = serialize_contiguously(Device, 1),
    IOList_____ = [ <<?_PCAP_OPEN_LIVE,
            DeviceLen:32/integer>>,             % I/O list length
          DeviceBinOrList,
          <<
          Snaplen:32/integer, 
          Promisc:32/integer, 
          To_Ms:32/integer
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

pcap_open_offline(Port,
     Fname
        ) when port(Port) -> % TODO: Add additional constraints here
    {FnameBinOrList, FnameLen} = serialize_contiguously(Fname, 1),
    IOList_____ = [ <<?_PCAP_OPEN_OFFLINE,
            FnameLen:32/integer>>,              % I/O list length
          FnameBinOrList,
          <<
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

my_set_immediate(Port,
     P, 
     Immediate
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_pcap_t, PIndex} = P,
    IOList_____ = <<?_MY_SET_IMMEDIATE,
            PIndex:32/integer, 
          Immediate:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

my_set_see_sent(Port,
     P, 
     See_Sent
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_pcap_t, PIndex} = P,
    IOList_____ = <<?_MY_SET_SEE_SENT,
            PIndex:32/integer, 
          See_Sent:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

pcap_setnonblock(Port,
     P, 
     Nonblock
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_pcap_t, PIndex} = P,
    IOList_____ = <<?_PCAP_SETNONBLOCK,
            PIndex:32/integer, 
          Nonblock:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

pcap_getnonblock(Port,
     P
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_pcap_t, PIndex} = P,
    IOList_____ = <<?_PCAP_GETNONBLOCK,
            PIndex:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

pcap_findalldevs(Port
        ) when port(Port) -> % TODO: Add additional constraints here
    IOList_____ = <<?_PCAP_FINDALLDEVS>>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

pcap_lookupdev(Port
        ) when port(Port) -> % TODO: Add additional constraints here
    IOList_____ = <<?_PCAP_LOOKUPDEV>>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

pcap_lookupnet(Port,
     Device
        ) when port(Port) -> % TODO: Add additional constraints here
    {DeviceBinOrList, DeviceLen} = serialize_contiguously(Device, 1),
    IOList_____ = [ <<?_PCAP_LOOKUPNET,
            DeviceLen:32/integer>>,             % I/O list length
          DeviceBinOrList,
          <<
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

my_pcap_compile(Port,
     P, 
     Str, 
     Optimize, 
     Netmask
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_pcap_t, PIndex} = P,
    {StrBinOrList, StrLen} = serialize_contiguously(Str, 1),
    IOList_____ = [ <<?_MY_PCAP_COMPILE,
            PIndex:32/integer, 
            StrLen:32/integer>>,                % I/O list length
          StrBinOrList,
          <<
          Optimize:32/integer, 
          Netmask:32/unsigned-integer
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

pcap_setfilter(Port,
     P, 
     Fp
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_pcap_t, PIndex} = P,
    {valmap_bpf_program, FpIndex} = Fp,
    IOList_____ = <<?_PCAP_SETFILTER,
            PIndex:32/integer, 
            FpIndex:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

my_pcap_freecode(Port,
     Fp
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_bpf_program, FpIndex} = Fp,
    IOList_____ = <<?_MY_PCAP_FREECODE,
            FpIndex:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

pcap_next(Port,
     P
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_pcap_t, PIndex} = P,
    IOList_____ = <<?_PCAP_NEXT,
            PIndex:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

pcap_datalink(Port,
     P
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_pcap_t, PIndex} = P,
    IOList_____ = <<?_PCAP_DATALINK,
            PIndex:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

pcap_snapshot(Port,
     P
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_pcap_t, PIndex} = P,
    IOList_____ = <<?_PCAP_SNAPSHOT,
            PIndex:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

pcap_is_swapped(Port,
     P
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_pcap_t, PIndex} = P,
    IOList_____ = <<?_PCAP_IS_SWAPPED,
            PIndex:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

pcap_major_version(Port,
     P
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_pcap_t, PIndex} = P,
    IOList_____ = <<?_PCAP_MAJOR_VERSION,
            PIndex:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

pcap_minor_version(Port,
     P
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_pcap_t, PIndex} = P,
    IOList_____ = <<?_PCAP_MINOR_VERSION,
            PIndex:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

pcap_stats(Port,
     P
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_pcap_t, PIndex} = P,
    IOList_____ = <<?_PCAP_STATS,
            PIndex:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

pcap_fileno(Port,
     P
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_pcap_t, PIndex} = P,
    IOList_____ = <<?_PCAP_FILENO,
            PIndex:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

pcap_geterr(Port,
     P
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_pcap_t, PIndex} = P,
    IOList_____ = <<?_PCAP_GETERR,
            PIndex:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

pcap_close(Port,
     P
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_pcap_t, PIndex} = P,
    IOList_____ = <<?_PCAP_CLOSE,
            PIndex:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

    
%%%
%%% Internal functions.
%%%

load_path(File) ->
    case lists:filter(fun(D) ->
                              case file:read_file_info(D ++ "/" ++ File) of
                                  {ok, _} -> true;
                                  _ -> false
                              end
                      end, code:get_path()) of
        [Dir|_] ->
            {ok, Dir};
        [] ->
            io:format("Error: ~s not found in code path\n", [File]),
            {error, enoent}
    end.

%%%
%%% Note that an 'xtra_return' that only returns one item in its
%%% tuple will return {Port, ok, {Thingie}}, so we'll return
%%% {ok, {Thingie}}, which is *sooooooo* maddening because I keep
%%% forgetting the extra tuple wrapper.  So, if there's only one
%%% thingie in the return tuple, strip it off: {ok, Thingie}
%%%

get_port_reply(Port) when port(Port) ->
    receive
        {Port, ok} = T -> proc_reply(T);
        {Port, ok, {M}} = T -> proc_reply(T);
        {Port, ok, M} = T -> proc_reply(T);
        {Port, error, {Reason}} = T -> proc_reply(T);
        {Port, error, Reason} = T -> proc_reply(T);
        %% Pipe driver messages
        {Port, {data, Bytes}} -> proc_reply(pipedrv_deser(Port, Bytes));
        {'EXIT', Port, Reason} -> throw({port_error, Reason});  % XXX too drastic?
        {Port, Reason} -> throw({port_error, Reason})   % XXX too drastic?
    end.

%% This function exists to provide consistency of replies 
%% given by linked-in and pipe drivers.  The "receive" statement
%% in get_port_reply/1 is specific because we want it to be
%% very selective about what it will grab out of the mailbox.
proc_reply({Port, ok}) when port(Port) ->
    ok;
proc_reply({Port, ok, {M}}) when port(Port) ->
    {ok, M};
proc_reply({Port, ok, M}) when port(Port) ->
    {ok, M};
proc_reply({Port, error, {Reason}}) when port(Port) ->
    {error, Reason};
proc_reply({Port, error, Reason}) when port(Port) ->
    {error, Reason}.


%%% io_list_len() is an extremely useful function.  BEAM has got this
%%% implemented quite efficiently in C.  It would be *fabulous* to be able
%%% to use it from Erlang via a BIF.

io_list_len(B) when binary(B) -> {B, size(B)};
io_list_len(L) when list(L) -> io_list_len(L, 0).
io_list_len([H|T], N) ->
    if
        H >= 0, H =< 255 -> io_list_len(T, N+1);
        list(H) -> io_list_len(T, io_list_len(H,N));
        binary(H) -> io_list_len(T, size(H) + N);
        true -> throw({error, partial_len, N})
    end;
io_list_len(H, N) when binary(H) -> 
    size(H) + N;
io_list_len([], N) -> 
    N.

%%% We need to make the binary thing we're passing in contiguous
%%% because the C function we're calling is expecting a single
%%% contiguous buffer.  If IOList is ["Hello, ", <<"World">>, "!"],
%%% that binary in the middle element will end up with the argument
%%% spanning three parts of an ErlIOVec.  If that happens, then we'd
%%% have to have the driver do the dirty work of putting the argument
%%% into a single contiguous buffer.
%%%
%%% Frankly, we're lazy, and this code is short and won't be much
%%% slower than doing it in C.

%%% 2nd arg: if 1, NUL-terminate the IOList

serialize_contiguously(B, 0) when binary(B) ->
    {B, size(B)};
serialize_contiguously([B], 0) when binary(B) ->
    {B, size(B)};
serialize_contiguously(IOList, 1) ->
    serialize_contiguously([IOList, 0], 0);
serialize_contiguously(IOList, 0) ->
    B = list_to_binary(IOList),
    {B, size(B)}.


%% pipedrv_deser/2 -- Deserialize the term that the pipe driver is
%% is returning to Erlang.  The pipe driver doesn't know it's a pipe
%% driver, it thinks it's a linked-in driver, so it tries to return
%% an arbitrary Erlang term to us.  The pipe-main program is sneaky:
%% it has a driver_output_term() function that serializes the term
%% that the driver built.  With the help of a list-as-stack, we
%% deserialize that term.

pipedrv_deser(Port, B) ->
    pipedrv_deser(Port, B, []).

pipedrv_deser(Port, <<>>, []) ->
    throw(icky_i_think);
pipedrv_deser(Port, <<>>, [T]) ->
    T;
pipedrv_deser(Port, <<?T_NIL:8, Rest/binary>>, Stack) ->
    pipedrv_deser(Port, Rest, [foo___foo_nil___|Stack]);
pipedrv_deser(Port, <<?T_ATOM:8, Len:8, Rest/binary>>, Stack) ->
    <<A:Len/binary, Rest2/binary>> = Rest,
    pipedrv_deser(Port, Rest2, [list_to_atom(binary_to_list(A))|Stack]);
pipedrv_deser(Port, <<?T_PORT:8, P:32/unsigned, Rest/binary>>, Stack) ->
    %% The pipe driver tried sending us a port, but it cannot know what
    %% port ID was assigned to this port, so we'll assume it is Port.
    pipedrv_deser(Port, Rest, [Port|Stack]);
pipedrv_deser(Port, <<?T_INT:8, I:32/signed, Rest/binary>>, Stack) ->
    pipedrv_deser(Port, Rest, [I|Stack]);
pipedrv_deser(Port, <<?T_TUPLE:8, N:8, Rest/binary>>, Stack) ->
    {L, NewStack} = popN(N, Stack),
    pipedrv_deser(Port, Rest, [list_to_tuple(L)|NewStack]);
pipedrv_deser(Port, <<?T_LIST:8, N:32, Rest/binary>>, Stack) ->
    {L, NewStack} = popN(N, Stack),
    pipedrv_deser(Port, Rest, [L|NewStack]);
pipedrv_deser(Port, <<?T_BINARY:8, Len:32/signed, Rest/binary>>, Stack) ->
    <<Bin:Len/binary, Rest2/binary>> = Rest,
    pipedrv_deser(Port, Rest2, [Bin|Stack]);
pipedrv_deser(Port, <<?T_STRING:8, Len:32/signed, Rest/binary>>, Stack) ->
    <<Bin:Len/binary, Rest2/binary>> = Rest,
    pipedrv_deser(Port, Rest2, [binary_to_list(Bin)|Stack]);
pipedrv_deser(Port, X, Y) ->
    throw({bah, X, Y}).

popN(N, Stack) ->
    popN(N, Stack, []).
popN(0, Stack, Acc) ->
    {Acc, Stack};
popN(N, [foo___foo_nil___|T], Acc) ->
    %% This is the nonsense we put on the stack to represent NIL.  Ignore it.
    popN(N - 1, T, Acc);
popN(N, [H|T], Acc) ->
    popN(N - 1, T, [H|Acc]).


%%%
%%% Begin code included via <custom_erl> tags
%%%


